:- use_module(list_length).

msort(L, S) :-
	list_length(L, D1, D),
	list_length(S, D2, D),
	msort_aux(D1, D2, D),
	msort_sdr(L, S, D).

:- block msort_aux(-, -, ?).
msort_aux(D1, D2, D) :-
	D1 = D2,
	(D1 = 0 ->
		D = 0;
		D is integer(ceiling(log(D1) / log(2)))
	).

:- block msort_sdr(?, ?, -).
msort_sdr([], [], _).
msort_sdr([X], [X], _).
msort_sdr([X, Y | Xs], S, D) :-
	D >= 1,
	D1 is D - 1,
	split([X, Y | Xs], L1, L2), 
	msort_sdr(L1, S1, D1), 
	msort_sdr(L2, S2, D1),
	merge(S1, S2, S).

split(L, L1, L2) :-
	list_length(L, S, D),
	list_length(L1, S1, D),
	list_length(L2, S2, D),
	split_aux(S, S1, S2, D),
	split_sdr(L, L1, L2, D).

:- block split_aux(-, -, -, ?).
split_aux(S, _, _, D) :-
	integer(S),
	!,
	D = S.
split_aux(_, S1, _, D) :-
	integer(S1),
	!,
	D is 2 * S1.
split_aux(_, _, S2, D) :-
	D is 2 * S2 + 1.

:- block split_sdr(?, ?, ?, -).
split_sdr([], [], [], _).
split_sdr([X | L], [X | L1], L2, D) :- 
	D >= 1,
	D1 is D - 1,
	split_sdr(L, L2, L1, D1).

merge(L1, L2, L) :-
	list_length(L1, S1, D),
	list_length(L2, S2, D),
	list_length(L, S, D),
	merge_aux(S1, S2, S, D),
	merge_sdr(L1, L2, L, D).

:- block merge_aux(-, ?, -, ?), merge_aux(?, -, -, ?).
merge_aux(S1, S2, _, D) :-
	integer(S1),
	integer(S2),
	!,
	D is S1 + S2 - 1.
merge_aux(_, _, S, D) :-
	D is S - 1.

:- block merge_sdr(?, ?, ?, -).
merge_sdr([], Ys, Ys, _).
merge_sdr(Xs, [], Xs, _).
merge_sdr([X | Xs], [Y | Ys], [X | Zs], D) :-
	D >= 1,
	D1 is D - 1,
	leq(X, Y), 			% X =< Y,
	merge_sdr(Xs, [Y | Ys], Zs, D1). 
merge_sdr([X | Xs], [Y | Ys], [Y | Zs], D) :-
	D >= 1,
	D1 is D - 1,
	gt(X, Y), 			% X > Y,
	merge_sdr([X | Xs], Ys, Zs, D1).

:- block leq(-, ?), leq(?, -).
leq(X, Y) :- X =< Y.

:- block gt(-, ?), gt(?, -).
gt(X, Y) :- X > Y.