:- use_module(qsort_control, [qsort/2]).
:- use_module(list_length).

qsort(L, S) :- qsort(L, S, []).

qsort(L, S, T) :-
	list_length(L, D1, D),
	list_length(S, D2, D),
	qsort_aux(D1, D2, D),
	qsort_sdr(L, S, T, D).

:- block qsort_aux(-, -, ?).
qsort_aux(D1, D2, D) :-
	D1 = D2, D = D1.

:- block qsort_sdr(?, ?, ?, -).
qsort_sdr([], X, X, _).
qsort_sdr([M | Xs], S1, S, D) :-
	1 =< D, 
	D1 is D - 1,
	part(Xs, M, L1, L2), 
	qsort_sdr(L1, S1, [M | S2], D1), 
	qsort_sdr(L2, S2, S, D1).

part(L, M, L1, L2) :-
	list_length(L, D1, D),
	list_length(L1, D2, D),
	list_length(L2, D3, D),
	part_aux(D1, D2, D3, D),
	part_sdr(L, M, L1, L2, D).

:- block part_aux(-, -, ?, ?), part_aux(-, ?, -, ?).
part_aux(D1, _, _, D) :- nonvar(D1), !, D = D1.
part_aux(_, D2, D3, D) :- D is D2 + D3.

:- block part_sdr(?, ?, ?, ?, -). 
part_sdr([], _, [], [], _).
part_sdr([X | L], M, L1, [X | L2], D) :-
	1 =< D,
	D1 is D - 1,
	leq(M, X), 
	part_sdr(L, M, L1, L2, D1).
part_sdr([X | L], M, [X | L1], L2, D) :-
	1 =< D,
	D1 is D - 1,
	gt(M, X), 
	part_sdr(L, M, L1, L2, D1).

:- block leq(-, ?), leq(?, -).
leq(X, Y) :- X =< Y.

:- block gt(-, ?), gt(?, -).
gt(X, Y) :- X > Y.