%:- use_module(queens_control, [main/2]).
:- use_module(list_length).

main(N, Time) :-
	statistics(walltime, _),
	findall(a, nqueens(N, _), _),
	statistics(walltime, [_, Time]).

nqueens(N, Soln):-
   	length(Range, N),
	length(Soln, N),
   	range(Range, 1),
	safe(Soln),		% test and generate version
	perm(Range, Soln).

range(L, N) :-
	list_length(L, D, _),
	range_sdr(L, N, D).	

:- block range_sdr(?, ?, -).
range_sdr([], _, _).
range_sdr([N | Ns], N, D) :- 
	D >= 1, D1 is D - 1, 
	N1 is N + 1, 
	range_sdr(Ns, N1, D1).

perm(L, P) :-
	list_length(L, D1, D),
	list_length(P, D2, D),
	perm_aux(D1, D2, D),
	perm_sdr(L, P, D).	

:- block perm_aux(-, ?, ?). perm_aux(?, -, ?).
perm_aux(D1, D2, D) :- D1 = D2, D = D1.

:- block perm_sdr(?, ?, -).
perm_sdr([], [], _).
perm_sdr([X | Xs], P, D) :-
	D >= 1, D1 is D - 1, 	
	select(P, X, Rs),
	perm_sdr(Xs, Rs, D1).

select(L, N, R) :-
	list_length(L, D1, D),
	list_length(R, D2, D),
	select_aux(D1, D2, D),
	select_sdr(L, N, R, D).

:- block select_aux(-, -, ?).
select_aux(D1, _, D) :- nonvar(D1), !, D = D1.
select_aux(_, D2, D) :- D is D2 + 1.

:- block select_sdr(?, ?, ?, -).
select_sdr([N | Ns], N, Ns, _).
select_sdr([N | Ns], S, [N | Rs], D) :- 
	D >= 1, D1 is D - 1, select_sdr(Ns, S, Rs, D1).

safe(L) :- 
	list_length(L, D, _), 
	safe_sdr(L, D).

:- block safe_sdr(?, -).
safe_sdr([], _).
safe_sdr([CN | CNs], D) :- D >= 1, D1 is D - 1,
	no_attack(CNs, CN, 1), safe_sdr(CNs, D1).

no_attack(CNs, CN, Diff) :- 
	list_length(CNs, D, _), 
	no_attack_sdr(CNs, CN, Diff, D). 

:- block no_attack_sdr(-, ?, ?).
no_attack_sdr([], _, _, _).
no_attack_sdr([CN|CNs], First_CN, Diff, D) :-
	D >= 1, D1 is D - 1,
	no_take(Diff, First_CN, CN),
   	Next_Diff is Diff + 1,
   	no_attack_sdr(CNs, First_CN, Next_Diff, D1).

:- block no_take(?, -, ?), no_take(?, ?, -).
no_take(Diff, First_CN, CN) :- 
	Diff =\= abs(First_CN - CN).