%:- use_module(queens_test_and_generate, [main/2]).

main(N, Time) :-
	statistics(walltime, _),
	findall(a, nqueens(N, _), _),
	statistics(walltime, [_, Time]).

nqueens(N, Soln):-
   	length(Range, N),
   	range(Range, 1),
	safe(Soln),
	perm(Soln, Range).

range([], _).
range([N | Ns], N) :- N1 is N + 1, range(Ns, N1).

perm([], []).
perm([X | Xs], P) :-
	select(P, X, Rs),
	perm(Xs, Rs).

select([N | Ns], N, Ns).
select([N | Ns], S, [N | Rs]) :- select(Ns, S, Rs).

:- block safe(-).
safe([]).
safe([CN | CNs]) :-
	no_attack(CNs, CN, 1),	safe(CNs).

:- block no_attack(-, ?, ?).
no_attack([], _, _).
no_attack([CN|CNs], First_CN, Diff) :-
	no_take(Diff, First_CN, CN),
   	Next_Diff is Diff + 1,
   	no_attack(CNs, First_CN, Next_Diff).

:- block no_take(?, -, ?), no_take(?, ?, -).
no_take(Diff, First_CN, CN) :- 
	Diff =\= abs(First_CN - CN).
